from ._model_registry import MODEL_DICT, ModelUniversalRegistry, register_model
from .base import BaseAutoModel
from .topkpool import AutoTopkpool


# from .graph_saint import GraphSAINTAggregationModel
from .gcn import GCN, AutoGCN
from .graphsage import GraphSAGE, AutoSAGE
from .gat import GAT,AutoGAT
from .gin import AutoGIN
from .hetero.hgt import AutoHGT
from .hetero.han import AutoHAN
from .hetero.HeteroRGCN import AutoHeteroRGCN

__all__ = [
    "ModelUniversalRegistry",
    "register_model",
    "BaseAutoModel",
    "AutoTopkpool",
    # "GraphSAINTAggregationModel",
    "GCN",
    "AutoGCN",
    "GraphSAGE",
    "AutoSAGE",
    "GAT",
    "AutoGAT",
    "AutoGIN",
    "AutoHGT",
    "AutoHAN", 
    "AutoHeteroRGCN",
]



import os
import torch
import torch_geometric
import logging

import numpy as np
import typing as _typing
from tqdm import trange
from typing import Union, Tuple, Sequence, Type, Callable

import torch.nn.functional as F
import torch.utils.data

from torch.optim.lr_scheduler import (
    StepLR,
    MultiStepLR,
    ExponentialLR,
    ReduceLROnPlateau,
)

from .losses import NTXent_loss
from .utils import get_view_by_name

from autogl.module.model.encoders.base_encoder import AutoHomogeneousEncoderMaintainer

from ..base import BaseTrainer, EarlyStopping, _DummyModel
from ..evaluation import Evaluation, get_feval, Acc
from ...model import (
    BaseAutoModel,
    BaseEncoderMaintainer, 
    BaseDecoderMaintainer,
    EncoderUniversalRegistry,
    DecoderUniversalRegistry
)
from ....utils import get_logger
from ....datasets import utils

LOGGER = get_logger("contrastive trainer")

class BaseContrastiveTrainer(BaseTrainer):
    def __init__(
        self, 
        encoder: _typing.Union[BaseEncoderMaintainer, str, None], 
        decoder: _typing.Union[BaseDecoderMaintainer, str, None],
        decoder_node: _typing.Union[BaseDecoderMaintainer, None],
        num_features: _typing.Union[int, None] = None,
        num_graph_features: _typing.Union[int, None] = None,
        device: _typing.Union[torch.device, str] = "auto",
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: Union[str, Callable] = "NT_Xent",
        f_loss: Union[str, Callable] = "nll_loss",
        views_fn: _typing.Union[
            _typing.Sequence[_typing.Callable], None
        ] = None,
        aug_ratio: Union[float, Sequence[float]] = 0.2,
        graph_level: bool = True,
        node_level: bool = False,
        z_dim: _typing.Union[int, None] = None,
        z_node_dim: _typing.Union[int, None] = None,
        tau: int = 0.5,
        p_optim: Union[torch.optim.Optimizer, str] = "Adam",
        p_lr: float = 0.0001,
        p_lr_scheduler_type: str = None,
        p_weight_decay: int = 0,
        p_epoch: int = 100,
        p_early_stopping_round: int = 20,
        f_optim: Union[torch.optim.Optimizer, str] = "Adam",
        f_lr: float = 0.001,
        f_lr_scheduler_type: str = None,
        f_weight_decay: int = 0,
        f_epoch: int = 100,
        f_early_stopping_round: int = 20,
        model_path: _typing.Union[str, None] = "./models"
    ):
        """
        The basic trainer for self-supervised learning with contrastive method.
        Used to automatically train the self-supervised problems.
        Parameters
        ----------
        encoder: `BaseEncoderMaintainer`, `str` or None
            A graph encoder shared by all views.
        decoder: `BaseDecoderMaintainer`, `str` or None
            A decoder which can be understood as a projection head for graph-level representations.
            Only required if `graph_level` = True.
        decoder_node: `BaseDecoderMaintainer`, `str` or None
            A decoder which can be understood as a projection head for node-level representations.
            Only required if `node_level` = True.
        num_features: `int` or None, Optional
            The number of features in dataset.
        num_graph_features: `int` or None, Optional
            The number of graph level features in dataset.
        device: `torch.device` or `str`, Optional
            The device this trainer will use.
            When `device` = "auto", if GPU exists in the device and dependency is installed, 
            the trainer will give priority to GPU, otherwise CPU will be used
        feval: a sequence of `str` or a sequence of `Evaluation`, Optional
            The evaluation methods.
        loss: `str` or `Callable`, Optional
            The loss function or the learning objective of contrastive model.
        views_fn: a list of `Callable` or None, Optional
            List of functions or augmentation methods to generate views from give graphs.
        graph_level: `bool`, Optional
            Whether to include graph-level representations
        node_level: `bool`, Optional
            Whether to include node-level representations
        z_dim: `int`, Optional
            The dimension of graph-level representations
        z_node_dim: `int`, Optional
            The dimension of node-level representations
        tau: `int`, Optional
            The temperature parameter in NT_Xent loss. Only used when `loss` = "NT_Xent"
        model_path: `str` or None, Optional
            The directory to restore the saved model.
            If `model_path` = None, the model will not be saved.
        """
        assert (node_level or graph_level) is True
        assert isinstance(encoder, BaseEncoderMaintainer) or isinstance(encoder, str) or encoder is None
        self.loss = self._get_loss(loss)
        self.node_level = node_level
        self.graph_level = graph_level
        self.z_dim = z_dim
        self.z_node_dim = z_node_dim
        self._encoder = None
        # parameters that record the results
        self.valid_result = None
        self.valid_result_prob = None
        self.valid_score = None
        # TODO
        # do not support method with both node-level representation and graph-level representation
        # so the decoder will be either _decoder or _decoder_node, one of them
        self._decoder = None
        self.views_fn_opt = views_fn
        self._views_fn = None
        self._aug_ratio = aug_ratio
        self.last_dim = z_dim if graph_level else z_node_dim
        self.num_features = num_features
        self.num_graph_features = num_graph_features
        self.tau = tau
        self.model_path = model_path
        if isinstance(device, str):
            if device == "auto":
                self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            else:
                self.device = torch.device(device)
        elif isinstance(device, torch.device):
            self.device = device
        else:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.p_opt_received = p_optim
        self.p_optimizer = self._get_optimizer(p_optim)
        self.p_lr_scheduler_type = p_lr_scheduler_type
        self.p_lr = p_lr
        self.p_epoch = p_epoch
        self.p_weight_decay = p_weight_decay
        self.p_early_stopping_round = (
            p_early_stopping_round if p_early_stopping_round is not None else 100
        )
        self.p_early_stopping = EarlyStopping(
            patience=self.p_early_stopping_round, verbose=False
        )
        self.f_opt_received = f_optim
        self.f_optimizer = self._get_optimizer(f_optim)
        self.f_lr_scheduler_type = f_lr_scheduler_type
        self.f_lr = f_lr
        self.f_epoch = f_epoch
        self.f_weight_decay = f_weight_decay
        self.f_early_stopping_round = (
            f_early_stopping_round if f_early_stopping_round is not None else 100
        )
        self.f_early_stopping = EarlyStopping(
            patience=self.f_early_stopping_round, verbose=False
        )
        if isinstance(f_loss, str) and hasattr(F, f_loss):
            self.f_loss = getattr(F, f_loss)
        elif callable(f_loss):
            self.f_loss = f_loss
        else:
            raise NotImplementedError(f"The loss {f_loss} is not supported yet.")
        super().__init__(
            encoder=encoder,
            decoder=decoder if graph_level else decoder_node,
            device=self.device,
            feval=feval,
            loss=self.loss
        )

    def _get_loss(self, loss):
        if callable(loss):
            return loss
        elif isinstance(loss, str):
            assert loss in ['NT_Xent']
            return {'NT_Xent': NTXent_loss}[loss]
        else:
            raise NotImplementedError("The argument `loss` should be str or callable which returns a loss tensor")

    # override encoder and decoder to depend on contrastive learning
    @property
    def encoder(self):
        return self._encoder
    
    @encoder.setter
    def encoder(self, enc: _typing.Union[BaseEncoderMaintainer, str, None]):
        if isinstance(enc, str):
            if enc in EncoderUniversalRegistry:
                if self.node_level:
                    self._encoder = EncoderUniversalRegistry.get_encoder(enc)(
                        self.num_features, final_dimension=self.last_dim, device=self.device, init=self.initialized
                    )
                elif self.graph_level:
                    self._encoder = EncoderUniversalRegistry.get_encoder(enc)(
                        self.num_features,
                        final_dimension=self.last_dim,
                        num_graph_featues=self.num_graph_features,
                        device=self.device,
                        init=self.initialized
                    )
            else:
                raise NotImplementedError(f"Sorry. Encoder {enc} is not supported yet.")
        elif isinstance(enc, BaseEncoderMaintainer):
            self._encoder = enc
        elif enc is None:
            self._encoder = None
        else:
            raise NotImplementedError(f"Sorry. Encoder {enc} is not supported yet.")
        self.num_features = self.num_features
        self.last_dim = self.last_dim
        self.num_graph_features = self.num_graph_features
    
    @property
    def decoder(self):
        if isinstance(self.encoder, BaseAutoModel): return None
        return self._decoder

    @decoder.setter
    def decoder(self, dec: _typing.Union[BaseDecoderMaintainer, str, None]):
        if isinstance(self.encoder, BaseAutoModel):
            logging.warn("Ignore passed dec since enc is a whole model")
            self._decoder = None
            return
        if isinstance(dec, str):
            if self.node_level:
                self._decoder = DecoderUniversalRegistry.get_decoder(dec)(
                    self.last_dim,
                    input_dimension=self.last_dim,
                    device=self.device,
                    init=self.initialized
                )
            elif self.graph_level:
                self._decoder = DecoderUniversalRegistry.get_decoder(dec)(
                    self.last_dim,
                    input_dimension=self.last_dim,
                    num_graph_features=self.num_graph_features,
                    device=self.device,
                    init=self.initialized
                )
        elif isinstance(dec, BaseDecoderMaintainer) or dec is None:
            self._decoder = dec
        else:
            raise NotImplementedError(f"Sorry. The decoder {dec} is not supported yet.")
        self.num_features = self.num_features
        self.last_dim = self.last_dim
        self.num_graph_features = self.num_graph_features

    @property
    def num_graph_features(self):
        return self._num_graph_features

    @num_graph_features.setter
    def num_graph_features(self, num_graph_featues):
        self._num_graph_features = num_graph_featues
        if self.graph_level:
            if self.encoder is not None: self.encoder.num_graph_features = self._num_graph_features
            if self.decoder is not None: self.decoder.num_graph_features = self._num_graph_features

    @property
    def num_features(self):
        return self._num_features

    @num_features.setter
    def num_features(self, num_features):
        self._num_features = num_features
        if self.encoder is not None:
            self.encoder.input_dimension = num_features

    @property
    def num_classes(self):
        return self._num_classes
    
    @num_classes.setter
    def num_classes(self, num_classes):
        self._num_classes = num_classes
        if self.prediction_head is not None:
            self.prediction_head.output_dimension = num_classes

    @property
    def views_fn(self):
        return self._views_fn
    
    @views_fn.setter
    def views_fn(self, views_fn):
        self.views_fn_opt = views_fn
        # set augmentation methods
        if isinstance(views_fn, list):
            if isinstance(self.aug_ratio, float):
                self.aug_ratio = [self.aug_ratio] * len(views_fn)
            if isinstance(self.aug_ratio, list) and len(self.aug_ratio) != len(views_fn):
                self.aug_ratio = [self.aug_ratio[0]] * len(views_fn)
        self._views_fn = self._get_views_fn(views_fn, self.aug_ratio)

    @property
    def aug_ratio(self):
        return self._aug_ratio
    
    @aug_ratio.setter
    def aug_ratio(self, aug_ratio):
        # set augmentation methods
        if isinstance(self.views_fn, list):
            if isinstance(aug_ratio, float):
                aug_ratio = [aug_ratio] * len(self.views_fn)
            assert len(aug_ratio) >= len(self.views_fn)
        self._aug_ratio = aug_ratio
        self._views_fn = self._get_views_fn(self.views_fn_opt, self.aug_ratio)

    @property
    def prediction_head(self):
        return self._prediction_head
    
    @prediction_head.setter
    def prediction_head(self, head: _typing.Union[BaseDecoderMaintainer, str, None]):
        if isinstance(self.encoder, BaseAutoModel):
            raise ValueError("Encoder shouldn't be a `BaseAutoModel` in GraphCLSemisupervisedTrainer.")
        if isinstance(head, str):
            self._prediction_head = DecoderUniversalRegistry.get_decoder(head)(
                self.num_classes,
                input_dim=self.last_dim,
                num_graph_features=self.num_graph_features,
                device=self.device,
                init=self.initialized
            )
        elif isinstance(head, BaseDecoderMaintainer) or head is None or (hasattr(head, 'fit') and hasattr(head, 'predict')):
            self._prediction_head = head
        else:
            raise NotImplementedError(f"Sorry. The head {head} is not supported yet.")
        self.num_features = self.num_features
        self.last_dim = self.last_dim
        self.num_graph_features = self.num_graph_features
        self.num_classes = self.num_classes

    def train(self, dataset, keep_valid_result=True):
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ----------
        dataset: The graph dataset used to be trained.
        keep_valid_result: `bool`
            if True, save the validation result after training.

        Returns
        -------
        self: `autogl.train.GraphCLSemisupervisedTrainer`
            A reference of current trainer.

        """
        try:
            if self.graph_level:
                valid_loader = utils.graph_get_split(
                    dataset, "val", batch_size=self.batch_size, num_workers=self.num_workers
                )
            else:
                # TODO node_level
                valid_loader = None
        except ValueError:
            valid_loader = None
        except AttributeError:
            valid_loader = None
        self._train_only(dataset)
        if keep_valid_result and valid_loader:
            # save the validation result after training
            pred = self._predict_only(valid_loader)
            self.valid_result = pred.max(1)[1]
            self.valid_result_prob = pred
            self.valid_score = self.evaluate(dataset, mask="val", feval=self.feval)

    def _get_optimizer(self, optimizer):
        if isinstance(optimizer, str):
            if optimizer.lower() == "adam": 
                optimizer = torch.optim.Adam
            elif optimizer.lower() == "sgd": 
                optimizer = torch.optim.SGD
            else: 
                raise ValueError("Currently not support optimizer {}".format(optimizer))
        elif isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            optimizer = optimizer
        else:
            raise ValueError("Currently not support optimizer {}".format(optimizer))
        return optimizer

    def _get_views_fn(self, views_fn, aug_ratio):
        # GraphCL only need two kinds of augmentation methods
        if views_fn is None:
            return None
        final_views_fn = []
        for i, view in enumerate(views_fn):
            if isinstance(view, str):
                assert view in ["dropN", "permE", "subgraph", "maskN", "random2", "random3", "random4"]
                final_views_fn.append(get_view_by_name(view, aug_ratio[i]))
            else:
                final_views_fn.append(view)
        return final_views_fn

    def _get_scheduler(self, stage, optimizer):
        if stage == 'pretraining':
            lr_scheduler_type = self.p_lr_scheduler_type
        else:
            lr_scheduler_type = self.f_lr_scheduler_type
        if type(lr_scheduler_type) == str and lr_scheduler_type == "steplr":
            scheduler = StepLR(optimizer, step_size=100, gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "multisteplr":
            scheduler = MultiStepLR(optimizer, milestones=[30, 80], gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "exponentiallr":
            scheduler = ExponentialLR(optimizer, gamma=0.1)
        elif (
            type(lr_scheduler_type) == str and lr_scheduler_type == "reducelronplateau"
        ):
            scheduler = ReduceLROnPlateau(optimizer, "min")
        else:
            scheduler = None
        return scheduler
    
    def _get_loader_loss(self, loader, optimizer=None, scheduler=None, mode="train"):
        epoch_loss = 0.0
        last_loss = 0.0
        if mode == "train":
            for data in loader:
                optimizer.zero_grad()
                if None in self.views_fn:
                    # For view fn that returns multiple views
                    views = []
                    for v_fn in self.views_fn:
                        if v_fn is not None:
                            views += [*v_fn(data)]
                else:
                    views = [v_fn(data) for v_fn in self.views_fn]
                zs = []
                for view in views:
                    z = self._get_embed(view.to(self.device))
                    zs.append(self.decoder.decoder(z, view.to(self.device)))
                loss = self.loss(zs, tau=self.tau)
                loss.backward()
                optimizer.step()
                if self.p_lr_scheduler_type:
                    scheduler.step()
                epoch_loss += loss.item()
                last_loss = loss.item()
        else:
            for data in loader:
                if None in self.views_fn:
                    # For view fn that returns multiple views
                    views = []
                    for v_fn in self.views_fn:
                        if v_fn is not None:
                            views += [*v_fn(data)]
                else:
                    views = [v_fn(data) for v_fn in self.views_fn]
                zs = []
                for view in views:
                    z = self._get_embed(view.to(self.device))
                    zs.append(self.decoder.decoder(z, view.to(self.device)))
                loss = self.loss(zs, tau=self.tau)
                epoch_loss += loss.item()
                last_loss = loss.item()
        return epoch_loss, last_loss

    def _train_pretraining_only(self, dataset, per_epoch=False):
        """
        Pretraining stage
        As a matter of fact, it trains encoder, and decoder is just an auxiliary task
        """
        import torch_geometric
        if int(torch_geometric.__version__.split('.')[0]) >= 2:
            # version 2.x
            from torch_geometric.loader import DataLoader
        else:
            from torch_geometric.data import DataLoader
        if self.graph_level:
            pre_train_loader = DataLoader(
                dataset, batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True
            )
        else:
            # TODO node level
            pre_train_loader = None
        optimizer = self.p_optimizer(
            self.encoder.encoder.parameters(), lr=self.p_lr, weight_decay=self.p_weight_decay
        )
        optimizer.add_param_group({"params": self.decoder.decoder.parameters()})
        scheduler = self._get_scheduler("pretraining", optimizer)
        with trange(self.p_epoch) as t:
            for epoch in t:
                self.encoder.encoder.train()
                self.decoder.decoder.train()
                t.set_description(f"Pretraining: epoch {epoch + 1}")
                try:
                    epoch_loss, last_loss = self._get_loader_loss(pre_train_loader, optimizer, scheduler, "train")
                except ValueError:
                    epoch_loss = 1e10
                    pass
                t.set_postfix(loss="{:.6f}".format(float(last_loss)))
                if per_epoch:
                    yield epoch
                else:
                    self.p_early_stopping(epoch_loss, self.encoder.encoder)
                    if self.p_early_stopping.early_stop:
                        LOGGER.debug("Early stopping at", epoch)
                        break
        
        if not per_epoch:
            self.p_early_stopping.load_checkpoint(self.encoder.encoder)
            yield self.p_epoch

    def _compose_model(self, pretrain=False):
        if pretrain:
            return _DummyModel(self.encoder, self.decoder).to(self.device)
        elif self.prediction_head is not None and isinstance(self.prediction_head, BaseDecoderMaintainer):
            return _DummyModel(self.encoder, self.prediction_head).to(self.device)
        else:
            return self.encoder.encoder.to(self.device)

    def _get_embed(self, view):
        z = self.encoder.encoder(view)
        return z

    def predict(self, dataset, mask="test"):
        """
        The function of predicting on the given dataset.

        Parameters
        ----------
        dataset: The graph classification dataset used to be predicted.

        mask: `str`
        "train", "val" or "test"
        The dataset mask.

        Returns
        -------
        The prediction result of `predict_proba`.
        """
        if self.graph_level:
            loader = utils.graph_get_split(
                dataset, mask, batch_size=self.batch_size, num_workers=self.num_workers
            )
        else:
            # TODO node level
            loader = None
        return self._predict_proba(loader, in_log_format=True).max(1)[1]
    
    def predict_proba(self, dataset, mask="test", in_log_format=False):
        """
        The function of predicting the probability on the given dataset.

        Parameters
        ----------
        dataset: The graph dataset used to be predicted.

        mask: `str`
        "train", "val" or "test"
        The dataset mask.

        in_log_format: `bool`
        If True(False), the probability will (not) be log format.

        Returns
        -------
        The prediction result.
        """
        if self.graph_level:
            loader = utils.graph_get_split(
                dataset, mask, batch_size=self.batch_size, num_workers=self.num_workers
            )
        else:
            # TODO node level
            loader = None
        return self._predict_proba(loader, in_log_format)

    def _predict_proba(self, loader, in_log_format=False, return_label=False):
        if return_label:
            ret, label = self._predict_only(loader, return_label=True)
        else:
            ret = self._predict_only(loader, return_label=False)
        if in_log_format is False:
            ret = torch.exp(ret)
        if return_label:
            return ret, label
        else:
            return ret

    def _predict_only(self, loader, return_label=False):
        raise NotImplementedError()

    def evaluate(self, dataset, mask="val", feval=None):
        """
        The function of evaluating the model on the given dataset and keeping valid result.

        Parameters
        ----------
        dataset: The graph dataset used to be evaluated.

        mask: `str`
        "Train", "val" or "test"
        The dataset mask

        feval: `str`
        The evaluation method used in this function

        Returns
        -------
        res: The evaluation result on the given dataset.
        """
        if self.graph_level:
            loader = utils.graph_get_split(
                dataset, mask, batch_size=self.batch_size, num_workers=self.num_workers
            )
        else:
            # TODO node level
            loader = None
        return self._evaluate(loader, feval)
    
    def _evaluate(self, loader, feval=None):
        if feval is None:
            feval = self.feval
        else:
            feval = get_feval(feval)
        
        y_pred_prob, y_true = self._predict_proba(loader=loader, return_label=True)
        y_pred = y_pred_prob.max(1)[1]

        if not isinstance(feval, list):
            feval = [feval]
            return_signle = True
        else:
            return_signle = False

        res = []
        for f in feval:
            flag = False
            try:
                res.append(f.evaluate(y_pred_prob, y_true))
                flag = False
            except:
                flag = True
            if flag:
                try:
                    res.append(
                        f.evaluate(y_pred_prob.cpu().numpy(), y_true.cpu().numpy())
                    )
                    flag = False
                except:
                    flag = True
            if flag:
                try:
                    res.append(
                        f.evaluate(
                            y_pred_prob.detach().numpy(), y_true.detach().numpy()
                        )
                    )
                    flag = False
                except:
                    flag = True
            if flag:
                try:
                    res.append(
                        f.evaluate(
                            y_pred_prob.cpu().detach().numpy(),
                            y_true.cpu().detach().numpy(),
                        )
                    )
                    flag = False
                except:
                    flag = True
            if flag:
                assert False

        if return_signle:
            return res[0]
        return res

    def combined_hyper_parameter_space(self):
        return {
            "trainer": self.hyper_parameter_space,
            "encoder": self.encoder.hyper_parameter_space,
            "decoder": [] if self.decoder is None else self.decoder.hyper_parameter_space,
            "prediction_head": [] if self.prediction_head is None else self.prediction_head.hyper_parameter_space
        }

    def get_valid_predict(self):
    # """Get the valid result."""
        return self.valid_result

    def get_valid_predict_proba(self):
        # """Get the valid result (prediction probability)."""
        return self.valid_result_prob

    def get_valid_score(self, return_major=True):
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, then return only consists of the major result.
            If False, then return consists of the all results.

        Returns
        -------
        result: The valid score in training stage.
        """
        if isinstance(self.feval, list):
            if return_major:
                return self.valid_score[0], self.feval[0].is_higher_better()
            else:
                return self.valid_score, [f.is_higher_better() for f in self.feval]
        else:
            return self.valid_score, self.feval.is_higher_better()

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "p_optimizer": self.p_optimizer,
                "p_learning_rate": self.p_lr,
                "p_max_epoch": self.p_epoch,
                "p_early_stopping_round": self.p_early_stopping_round,
                "encoder": repr(self.encoder),
                "decoder": repr(self.decoder)
            }
        )


import os
import torch
import dgl

# from autogl.data.graph import GeneralStaticGraphGenerator
from autogl.data.graph.utils import conversion as _conversion
from autogl.data import InMemoryStaticGraphSet
from ._dataset_registry import DatasetUniversalRegistry


@DatasetUniversalRegistry.register_dataset("cora")
class CoraDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.CoraGraphDataset(
            os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(CoraDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(CoraDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label'],
        #                 'train_mask': dgl_graph.ndata['train_mask'],
        #                 'val_mask': dgl_graph.ndata['val_mask'],
        #                 'test_mask': dgl_graph.ndata['test_mask']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("CiteSeer".lower())
class CiteSeerDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.CiteseerGraphDataset(
            os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(CiteSeerDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(CiteSeerDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label'],
        #                 'train_mask': dgl_graph.ndata['train_mask'],
        #                 'val_mask': dgl_graph.ndata['val_mask'],
        #                 'test_mask': dgl_graph.ndata['test_mask']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("PubMed".lower())
class PubMedDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.PubmedGraphDataset(
            os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(PubMedDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(PubMedDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label'],
        #                 'train_mask': dgl_graph.ndata['train_mask'],
        #                 'val_mask': dgl_graph.ndata['val_mask'],
        #                 'test_mask': dgl_graph.ndata['test_mask']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("reddit")
class RedditDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.RedditDataset(
            raw_dir=os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(RedditDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(RedditDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label'],
        #                 'train_mask': dgl_graph.ndata['train_mask'],
        #                 'val_mask': dgl_graph.ndata['val_mask'],
        #                 'test_mask': dgl_graph.ndata['test_mask']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("amazon_computers")
class AmazonComputersDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.AmazonCoBuyComputerDataset(
            raw_dir=os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(AmazonComputersDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(AmazonComputersDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("amazon_photo")
class AmazonPhotoDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.AmazonCoBuyPhotoDataset(
            raw_dir=os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(AmazonPhotoDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(AmazonPhotoDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("coauthor_physics")
class CoauthorPhysicsDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.CoauthorPhysicsDataset(
            raw_dir=os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(CoauthorPhysicsDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(CoauthorPhysicsDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("coauthor_cs")
class CoauthorCSDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.CoauthorCSDataset(
            raw_dir=os.path.join(path, '_dgl')
        )
        dgl_graph: dgl.DGLGraph = dgl_dataset[0]
        super(CoauthorCSDataset, self).__init__(
            [_conversion.dgl_graph_to_general_static_graph(dgl_graph)]
        )
        # super(CoauthorCSDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'feat': dgl_graph.ndata['feat'],
        #                 'label': dgl_graph.ndata['label']
        #             },
        #             torch.vstack(dgl_graph.edges())
        #         )
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("mutag")
class MUTAGDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "MUTAG", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(MUTAGDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(MUTAGDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("enzymes")
class ENZYMESDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.TUDataset(
            "ENZYMES", raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['node_attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['node_labels']
            del dgl_graph.ndata['node_attr']
            del dgl_graph.ndata['node_labels']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(ENZYMESDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(ENZYMESDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'node_labels': dgl_graph.ndata['node_labels'],
        #                 'node_attr': dgl_graph.ndata['node_attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': label}
        #         ) for (dgl_graph, label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("imdb-b")
@DatasetUniversalRegistry.register_dataset("imdb-binary")
@DatasetUniversalRegistry.register_dataset("IMDBb".upper())
@DatasetUniversalRegistry.register_dataset("IMDBBinary".upper())
@DatasetUniversalRegistry.register_dataset("IMDBBinary")
class IMDBBinaryDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "IMDBBINARY", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(IMDBBinaryDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(IMDBBinaryDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("imdb-m")
@DatasetUniversalRegistry.register_dataset("imdb-multi")
@DatasetUniversalRegistry.register_dataset("IMDBm".upper())
@DatasetUniversalRegistry.register_dataset("IMDBMulti".upper())
@DatasetUniversalRegistry.register_dataset("IMDBMulti")
class IMDBMultiDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "IMDBMULTI", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(IMDBMultiDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(IMDBMultiDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("reddit-b")
@DatasetUniversalRegistry.register_dataset("reddit-binary")
@DatasetUniversalRegistry.register_dataset("RedditB".upper())
@DatasetUniversalRegistry.register_dataset("RedditBinary".upper())
class RedditBinaryDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "REDDITBINARY", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(RedditBinaryDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(RedditBinaryDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("reddit-multi-5k")
@DatasetUniversalRegistry.register_dataset("RedditMulti5K".upper())
class REDDITMulti5KDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "REDDITMULTI5K", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(REDDITMulti5KDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )

        # super(REDDITMulti5KDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("collab")
class COLLABDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "COLLAB", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(COLLABDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(COLLABDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("proteins")
class ProteinsDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "PROTEINS", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(ProteinsDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(ProteinsDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("ptc-mr")
class PTCMRDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "PTC", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(PTCMRDataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(PTCMRDataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )


@DatasetUniversalRegistry.register_dataset("nci1")
class NCI1Dataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        dgl_dataset = dgl.data.GINDataset(
            "NCI1", False, raw_dir=os.path.join(path, "_dgl")
        )

        def _transform(dgl_graph: dgl.DGLGraph, label: torch.Tensor):
            dgl_graph.ndata['feat'] = dgl_graph.ndata['attr']
            dgl_graph.ndata['node_label'] = dgl_graph.ndata['label']
            del dgl_graph.ndata['attr']
            del dgl_graph.ndata['label']
            static_graph = _conversion.dgl_graph_to_general_static_graph(dgl_graph)
            static_graph.data['label'] = label
            return static_graph

        super(NCI1Dataset, self).__init__(
            [_transform(dgl_graph, label) for (dgl_graph, label) in dgl_dataset]
        )
        # super(NCI1Dataset, self).__init__(
        #     [
        #         GeneralStaticGraphGenerator.create_homogeneous_static_graph(
        #             {
        #                 'label': dgl_graph.ndata['label'],
        #                 'attr': dgl_graph.ndata['attr']
        #             },
        #             torch.vstack(dgl_graph.edges()),
        #             graph_data={'label': graph_label}
        #         )
        #         for (dgl_graph, graph_label) in dgl_dataset
        #     ]
        # )



from .rl import *
import numpy as np
class AGNNReinforceController(ReinforceController):
    def resample(self,search_fields,selection):
        # search_fields act as group of fields in the paper (like activation group)
        self._initialize()
        result = selection.copy()

        # 1. update initial state with fields not searched
        for field in self.fields:
            if field not in search_fields:
                self._update_state(field,selection[field.name])

        # 2. get probability of field to search
        for field in search_fields:
            result[field.name] = self._sample_single(field)
        return result

    def _update_state(self,field,sampled):
        # use unsearched fields as rnn history to update state
        self._lstm_next_step()
        self._inputs = self.embedding[field.name](torch.LongTensor([sampled]).to(self._inputs.device))

class AGNNActionGuider(nn.Module):
    def __init__(self, fields, groups,guide_type, **controllargs):
        super(AGNNActionGuider, self).__init__()
        # create independent controllers for each group 
        controllers=[AGNNReinforceController(fields,**controllargs) for group in groups]
        self.controllers=nn.ModuleList(controllers)
        self.fields=fields
        self.groups=groups
        self.guide_type = guide_type

    def dummy_selection(self):
        # create dummy selection 
        result=dict()
        for field in self.fields:
            result[field.name]=0
        return result

    def resample(self,selection):
        entropys=[]
        new_selections=[]
        sample_probs=[]
        for idx,cont in enumerate(self.controllers):
            cont=self.controllers[idx]
            group=self.groups[idx]
            new_selection=cont.resample(group,selection)
            new_selections.append(new_selection)
            entropy=cont.sample_entropy
            entropys.append(entropy)
            sample_probs.append(cont.sample_log_prob)
        print(f'$$entropys {entropys}')
        if self.guide_type==0:
            # use the most uncertain one 
            idx=np.argmax(entropys)
        elif self.guide_type==1:
            # or sample by using entropy 
            idx=torch.multinomial(F.softmax(torch.tensor(entropys),dim=0),1).item()
        else:
            assert False,f"Not implemented guide type {self.guide_type}"
        group=self.groups[idx]
        print(f'$$select group {group}')
        new_selection=new_selections[idx]
        self.sample_log_prob=sample_probs[idx]
        self.sample_entropy=entropys[idx]
        print(f'$$new selection {new_selection}')
        return new_selection

@register_nas_algo("agnn")
class AGNNRL(GraphNasRL):
    def __init__(self,guide_type=1,*args,**kwargs):
        super(AGNNRL, self).__init__(*args,**kwargs)
        self.guide_type = guide_type
    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset  # .to(self.device)
        self.estimator = estimator
        # replace choice
        self.nas_modules = []

        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)

        # to device
        self.model = self.model.to(self.device)
        # fields
        self.nas_fields = [
            ReinforceField(
                name,
                len(module),
                isinstance(module, PathSamplingLayerChoice) or module.n_chosen == 1,
            )
            for name, module in self.nas_modules
        ]

        # create groups
        tags='op in act concat'.split()
        groups={tag:[] for tag in tags}
        for field in self.nas_fields:
            for tag in tags:
                if tag in field.name:
                    groups[tag].append(field)
        groups=[x for x in groups.values() if x]

        # controller
        self.controller = AGNNActionGuider(
            self.nas_fields,
            groups,
            self.guide_type,
            lstm_size=100,
            temperature=5.0,
            tanh_constant=2.5,
            **(self.ctrl_kwargs or {}),
        )
        self.ctrl_optim = torch.optim.Adam(
            self.controller.parameters(), lr=self.ctrl_lr
        )

        # init selection (acc,selection)
        self.best_selection=[0,self.controller.dummy_selection()]

        # train
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                l2 = self._train_controller(i)
                bar.set_postfix(reward_controller=l2)

        selection=self.export()

        # selections = [x[1] for x in self.hist]
        # candidiate_accs = [-x[0] for x in self.hist]
        # # print('candidiate accuracies',candidiate_accs)
        # selection = self._choose_best(selections)
        arch = space.parse_model(selection, self.device)
        print(selection,arch)
        return arch

    def _train_controller(self, epoch):
        self.model.eval()
        self.controller.train()
        self.ctrl_optim.zero_grad()
        rewards = []
        selections=[]
        # baseline = None
        baseline=self.best_selection[0]
        # diff: graph nas train 100 and derive 100 for every epoch(10 epochs), we just train 100(20 epochs). totol num of samples are same (2000)
        with tqdm(
            range(self.ctrl_steps_aggregate), disable=self.disable_progress
        ) as bar:
            for ctrl_step in bar:
                self._resample()
                selections.append(self.selection.copy())
                metric, loss, hardware_metric = self._infer(mask="val")
                reward = metric

                # bar.set_postfix(acc=metric,loss=loss.item())
                LOGGER.debug(f"{self.arch}\n{self.selection}\n{metric},{loss}")
                # diff: not do reward shaping as in graphnas code
                if (
                    self.hardware_metric_limit is None
                    or hardware_metric[0] < self.hardware_metric_limit
                ):
                    self.hist.append([-metric, self.selection])
                    self.allhist.append([-metric, self.selection])
                    if len(self.hist) > self.topk:
                        self.hist.sort(key=lambda x: x[0])
                        self.hist.pop()
                rewards.append(reward)

                if self.entropy_weight:
                    reward += (
                        self.entropy_weight * self.controller.sample_entropy.item()
                    )

                if not baseline:
                    baseline = reward
                else:
                    baseline = baseline * self.baseline_decay + reward * (
                        1 - self.baseline_decay
                    )

                loss = self.controller.sample_log_prob * (reward - baseline)
                self.ctrl_optim.zero_grad()
                loss.backward()

                self.ctrl_optim.step()

                bar.set_postfix(acc=metric, max_acc=max(rewards))

            # conserative explorer: update the best selection
            idx=np.argmax(rewards)
            best_reward=rewards[idx]
            best_selection=selections[idx]
            if best_reward>self.best_selection[0]:
                self.best_selection=[best_reward,best_selection]

            print(f'$$best selection: {self.best_selection}')
        LOGGER.info("epoch:{}, mean rewards:{}".format(epoch, sum(rewards) / len(rewards)))
        return sum(rewards) / len(rewards)

    def _resample(self):
        result = self.controller.resample(self.best_selection[1])
        self.arch = self.model.parse_model(result, device=self.device)
        self.selection = result

    def export(self):
        # self.controller.eval()
        # with torch.no_grad():
        #     return self.controller.resample()
        return self.best_selection[1]



import torch
import torch.nn.functional as F
from torch_geometric.nn import GATConv
from . import register_model
from .base import BaseAutoModel, activate_func
from ....utils import get_logger

LOGGER = get_logger("GATModel")


def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args


class GAT(torch.nn.Module):
    def __init__(self, args):
        super(GAT, self).__init__()
        self.args = args
        self.num_layer = int(self.args["num_layers"])

        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_layers",
                    "hidden",
                    "heads",
                    "dropout",
                    "act",
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        if not self.num_layer == len(self.args["hidden"]) + 1:
            LOGGER.warn("Warning: layer size does not match the length of hidden units")
        self.convs = torch.nn.ModuleList()
        self.convs.append(
            GATConv(
                self.args["features_num"],
                self.args["hidden"][0],
                heads=self.args["heads"],
                dropout=self.args["dropout"],
            )
        )
        last_dim = self.args["hidden"][0] * self.args["heads"]
        for i in range(self.num_layer - 2):
            self.convs.append(
                GATConv(
                    last_dim,
                    self.args["hidden"][i + 1],
                    heads=self.args["heads"],
                    dropout=self.args["dropout"],
                )
            )
            last_dim = self.args["hidden"][i + 1] * self.args["heads"]
        self.convs.append(
            GATConv(
                last_dim,
                self.args["num_class"],
                heads=1,
                concat=False,
                dropout=self.args["dropout"],
            )
        )

    def forward(self, data):
        try:
            x = data.x
        except:
            print("no x")
            pass
        try:
            edge_index = data.edge_index
        except:
            print("no index")
            pass
        try:
            edge_weight = data.edge_weight
        except:
            edge_weight = None
            pass

        for i in range(self.num_layer):
            x = F.dropout(x, p=self.args["dropout"], training=self.training)
            x = self.convs[i](x, edge_index, edge_weight)
            if i != self.num_layer - 1:
                x = activate_func(x, self.args["act"])

        return F.log_softmax(x, dim=1)

    def lp_encode(self, data):
        x = data.x
        for i in range(self.num_layer - 1):
            x = self.convs[i](x, data.edge_index) # Jie
            if i != self.num_layer - 2:
                x = activate_func(x, self.args["act"])
                # x = F.dropout(x, p=self.args["dropout"], training=self.training)
        return x

    def lp_decode(self, z, pos_edge_index, neg_edge_index):
        edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
        logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
        return logits

    def lp_decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()


@register_model("gat-model")
class AutoGAT(BaseAutoModel):
    r"""
    AutoGAT. The model used in this automodel is GAT, i.e., the graph attentional network from the `"Graph Attention Networks"
    <https://arxiv.org/abs/1710.10903>`_ paper. The layer is

    .. math::
        \mathbf{x}^{\prime}_i = \alpha_{i,i}\mathbf{\Theta}\mathbf{x}_{i} +
        \sum_{j \in \mathcal{N}(i)} \alpha_{i,j}\mathbf{\Theta}\mathbf{x}_{j}

    where the attention coefficients :math:`\alpha_{i,j}` are computed as

    .. math::
        \alpha_{i,j} =
        \frac{
        \exp\left(\mathrm{LeakyReLU}\left(\mathbf{a}^{\top}
        [\mathbf{\Theta}\mathbf{x}_i \, \Vert \, \mathbf{\Theta}\mathbf{x}_j]
        \right)\right)}
        {\sum_{k \in \mathcal{N}(i) \cup \{ i \}}
        \exp\left(\mathrm{LeakyReLU}\left(\mathbf{a}^{\top}
        [\mathbf{\Theta}\mathbf{x}_i \, \Vert \, \mathbf{\Theta}\mathbf{x}_k]
        \right)\right)}.

    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    args: Other parameters.
    """

    def __init__(
        self, num_features=None, num_classes=None, device=None, **args
    ):
        super().__init__(num_features, num_classes, device, **args)
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "heads",
                "type": "DISCRETE",
                "feasiblePoints": "2,4,8,16",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

        self.hyper_parameters = {
            "num_layers": 2,
            "hidden": [32],
            "heads": 4,
            "dropout": 0.2,
            "act": "leaky_relu",
        }

    def _initialize(self):
        # """Initialize model."""
        self._model = GAT({
            "features_num": self.input_dimension,
            "num_class": self.output_dimension,
            **self.hyper_parameters
        }).to(self.device)



import numpy as np
import pandas as pd
import torch
import typing as _typing
import autogl
from autogl.data.graph import GeneralStaticGraph
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry
import lightgbm
from sklearn.model_selection import train_test_split
from ._basic import BaseFeatureSelector


def _gbdt_generator(
        data: autogl.data.Data, fixlen: int = 1000,
        params: _typing.Mapping[str, _typing.Any] = ...,
        is_val: bool = True, train_val_ratio: float = 0.2,
        **optimizer_parameters
) -> _typing.Optional[np.ndarray]:
    parameters: _typing.Dict[str, _typing.Any] = (
        dict(params)
        if (
                params not in (Ellipsis, None) and
                isinstance(params, _typing.Mapping)
        )
        else {
            "boosting_type": "gbdt",
            "verbosity": -1,
            "random_state": 47,
            "objective": "multiclass",
            "metric": ["multi_logloss"],
            "max_bin": 63,
            "save_binary": True,
            "num_threads": 20,
            "num_leaves": 16,
            "subsample": 0.9,
            "subsample_freq": 1,
            "colsample_bytree": 0.8,
            # 'is_training_metric': True,
            # 'metric_freq': 1,
        }
    )

    num_classes: int = torch.max(data.y).item() + 1
    parameters["num_class"] = num_classes
    __optimizer_parameters = {
        "num_boost_round": 100,
        "early_stopping_rounds": 5,
        "verbose_eval": False
    }
    __optimizer_parameters.update(optimizer_parameters)
    if hasattr(data, "train_mask") and data.train_mask is not None and (
            isinstance(data.train_mask, np.ndarray) or
            isinstance(data.train_mask, torch.Tensor)
    ):
        x: np.ndarray = data.x[data.train_mask].numpy()
        label: np.ndarray = data.y[data.train_mask].numpy()
    else:
        x: np.ndarray = data.x.numpy()
        label: np.ndarray = data.y.numpy()
        is_val: bool = False
    _, num_features = x.shape
    if num_features < fixlen:
        return None

    feature_index: np.ndarray = np.array(
        [f"f{i}" for i in range(num_features)]
    )
    if is_val:
        x_train, x_val, y_train, y_val = train_test_split(
            x, label, test_size=train_val_ratio, stratify=label, random_state=47
        )
        dtrain = lightgbm.Dataset(x_train, label=y_train)
        dval = lightgbm.Dataset(x_val, label=y_val)
        clf = lightgbm.train(
            train_set=dtrain, params=parameters, valid_sets=dval,
            **__optimizer_parameters
        )
    else:
        train_x = pd.DataFrame(x, columns=feature_index, index=None)
        dtrain = lightgbm.Dataset(train_x, label=label)
        clf = lightgbm.train(
            train_set=dtrain, params=parameters,
            **__optimizer_parameters
        )

    imp = np.array(list(clf.feature_importance()))
    return np.argsort(imp)[-fixlen:]


@DataPreprocessorUniversalRegistry.register_data_preprocessor("gbdt")
class GBDTFeatureSelector(BaseFeatureSelector):
    r"""simple wrapper of lightgbm , using importance ranking to select top-k features.

    Parameters
    ----------
    fixlen : int
        K for top-K important features.
    """

    def __init__(self, fixlen: int = 10, *args, **kwargs):
        super(GBDTFeatureSelector, self).__init__()
        self.__fixlen = fixlen
        self.__args = args
        self.__kwargs = kwargs

    def _fit(self, homogeneous_static_graph: GeneralStaticGraph) -> GeneralStaticGraph:
        if not isinstance(homogeneous_static_graph, GeneralStaticGraph):
            raise TypeError
        elif not (
            homogeneous_static_graph.nodes.is_homogeneous and
            homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError
        if 'x' in homogeneous_static_graph.nodes.data:
            features: torch.Tensor = homogeneous_static_graph.nodes.data['x']
        elif 'feat' in homogeneous_static_graph.nodes.data:
            features: torch.Tensor = homogeneous_static_graph.nodes.data['feat']
        else:
            raise ValueError("Node features not exists")
        if 'y' in homogeneous_static_graph.nodes.data:
            label: torch.Tensor = homogeneous_static_graph.nodes.data['y']
        elif 'label' in homogeneous_static_graph.nodes.data:
            label: torch.Tensor = homogeneous_static_graph.nodes.data['label']
        else:
            raise ValueError("Node label not exists")
        if 'train_mask' in homogeneous_static_graph.nodes.data:
            train_mask: _typing.Optional[torch.Tensor] = (
                homogeneous_static_graph.nodes.data['train_mask']
            )
        else:
            train_mask: _typing.Optional[torch.Tensor] = None
        data = autogl.data.Data(
            edge_index=homogeneous_static_graph.edges.connections,
            x=features, y=label
        )
        data.train_mask = train_mask
        self._selection = _gbdt_generator(
            data, self.__fixlen, *self.__args, **self.__kwargs
        )
        return homogeneous_static_graph



import torch
from typing import Union


def get_device(device: Union[str, torch.device]):
    """
    Get device of passed argument. Will return a torch.device based on passed arguments.
    Can parse auto, cpu, gpu, cpu:x, gpu:x, etc. If auto is given, will automatically find
    available devices.


    Parameters
    ----------
    device: ``str`` or ``torch.device``
        The device to parse. If ``auto`` if given, will determine automatically.

    Returns
    -------
    device: ``torch.device``
        The parsed device.
    """
    assert isinstance(
        device, (str, torch.device)
    ), "Only support device of str or torch.device, get {} instead".format(device)
    if device == "auto":
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    return torch.device(device)



import torch
import typing as _typing
import torch_geometric
import autogl
from autogl.data.graph import GeneralStaticGraph
from autogl.data.graph.utils.conversion import static_graph_to_pyg_data
from .train_test_split_edges import train_test_split_edges


def split_edges_for_data(
        data: _typing.Union[
            torch_geometric.data.Data, autogl.data.graph.GeneralStaticGraph, _typing.Any
        ],
        train_ratio: float, val_ratio: float
) -> torch_geometric.data.Data:
    if isinstance(data, torch_geometric.data.Data):
        if not (
                isinstance(data.edge_index, torch.Tensor) and
                data.edge_index.dim() == data.edge_index.size(0) == 2
        ):
            raise ValueError
        edge_index: torch.LongTensor = data.edge_index
        edge_attr: _typing.Optional[torch.Tensor] = data.edge_attr
        import copy
        __data = copy.copy(data)
    elif isinstance(data, autogl.data.graph.GeneralStaticGraph):
        if not (data.nodes.is_homogeneous and data.edges.is_homogeneous):
            raise ValueError(
                "Provided instance of GeneralStaticGraph MUST be homogeneous"
            )
        edge_index: torch.LongTensor = data.edges.connections
        edge_attr: _typing.Optional[torch.Tensor] = (
            data.edges.data['edge_attr'] if 'edge_attr' in data.edges.data else None
        )
        __data = static_graph_to_pyg_data(data)
    elif (
            hasattr(data, 'edge_index') and
            isinstance(data.edge_index, torch.Tensor) and
            data.edge_index.dim() == data.edge_index.size(0) == 2
    ):
        edge_index: torch.LongTensor = data.edge_index
        if (
                hasattr(data, 'edge_attr') and
                isinstance(data.edge_attr, torch.Tensor) and
                data.edge_attr.size(0) == edge_index.size(1)
        ):
            edge_attr: _typing.Optional[torch.Tensor] = data.edge_attr
        else:
            edge_attr: _typing.Optional[torch.Tensor] = None
        if hasattr(data, 'x') and isinstance(data.x, torch.Tensor):
            x: _typing.Optional[torch.Tensor] = data.x
        else:
            x: _typing.Optional[torch.Tensor] = None
        if hasattr(data, 'y') and isinstance(data.y, torch.Tensor):
            y: _typing.Optional[torch.Tensor] = data.y
        else:
            y: _typing.Optional[torch.Tensor] = None
        __data = torch_geometric.data.Data(
            edge_index=edge_index, edge_attr=edge_attr, x=x, y=y
        )
    else:
        raise ValueError

    if isinstance(val_ratio, float) and 0 < val_ratio < 1:
        test_ratio = 1 - train_ratio - val_ratio
    else:
        test_ratio = 1 - train_ratio
    compound_results = train_test_split_edges(
        edge_index, edge_attr,
        val_ratio=val_ratio, test_ratio=test_ratio
    )
    __data.train_pos_edge_index = compound_results.train_pos_edge_index
    __data.train_pos_edge_attr = compound_results.train_pos_edge_attr
    __data.train_neg_adj_mask = compound_results.train_neg_adj_mask
    __data.val_pos_edge_index = compound_results.val_pos_edge_index
    __data.val_pos_edge_attr = compound_results.val_pos_edge_attr
    __data.val_neg_edge_index = compound_results.val_neg_edge_index
    __data.test_pos_edge_index = compound_results.test_pos_edge_index
    __data.test_pos_edge_attr = compound_results.test_pos_edge_attr
    __data.test_neg_edge_index = compound_results.test_neg_edge_index
    return __data



import torch
import typing as _typing
from ...hpo import AutoModule
from ..encoders import base_encoder


class BaseDecoderMaintainer(AutoModule):
    def _initialize(
            self, encoder: base_encoder.AutoHomogeneousEncoderMaintainer, *args, **kwargs
    ) -> _typing.Optional[bool]:
        """ Abstract initialization method to override """
        raise NotImplementedError

    def from_hyper_parameter_and_encoder(
            self, hp: _typing.Mapping[str, _typing.Any],
            encoder: base_encoder.BaseEncoderMaintainer
    ) -> 'BaseDecoderMaintainer':
        duplicate = self.__class__(
            self.output_dimension, self.device
        )
        new_hp = dict(self.hyper_parameters)
        new_hp.update(hp)
        duplicate.hyper_parameters = new_hp
        duplicate.initialize(encoder)
        return duplicate

    def __init__(
            self, output_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(BaseDecoderMaintainer, self).__init__(
            device, *args, **kwargs
        )
        self.output_dimension = output_dimension
        self._decoder: _typing.Optional[torch.nn.Module] = None

    @property
    def output_dimension(self):
        return self.__output_dimension
    
    @output_dimension.setter
    def output_dimension(self, output_dimension):
        self.__output_dimension = output_dimension

    @property
    def decoder(self) -> _typing.Optional[torch.nn.Module]:
        return self._decoder

    def to_device(self, device: _typing.Union[torch.device, str, int, None]):
        self.device = device
        if (
                self._decoder not in (Ellipsis, None) and
                isinstance(self._decoder, torch.nn.Module)
        ):
            self._decoder.to(self.device)



from ....backend import DependentBackend

if DependentBackend.is_pyg():
    from .node_classification_sampled_trainer import (
        NodeClassificationGraphSAINTTrainer,
        NodeClassificationLayerDependentImportanceSamplingTrainer,
        NodeClassificationNeighborSamplingTrainer
    )
    __all__ = [
        "NodeClassificationGraphSAINTTrainer",
        "NodeClassificationLayerDependentImportanceSamplingTrainer",
        "NodeClassificationNeighborSamplingTrainer"
    ]
else:
    __all__ = []



import numpy as np
import scipy.sparse as sp
import typing as _typing
import dgl
import autogl.data.graph
from autogl.data.graph.utils.conversion import general_static_graph_to_dgl_graph


class _SplitEdgesDGLImpl:
    @classmethod
    def __split_edges_train_val_test(
            cls, g: dgl.DGLGraph,
            train_ratio: float, val_ratio: float
    ) -> _typing.Tuple[
        dgl.DGLGraph, dgl.DGLGraph, dgl.DGLGraph,
        dgl.DGLGraph, dgl.DGLGraph, dgl.DGLGraph, dgl.DGLGraph
    ]:
        u, v = g.edges()

        eids = np.arange(g.number_of_edges())
        eids = np.random.permutation(eids)

        valid_size = int(len(eids) * val_ratio)
        test_size = int(len(eids) * (1 - train_ratio - val_ratio))
        train_size = g.number_of_edges() - test_size - valid_size

        test_pos_u, test_pos_v = u[eids[:test_size]], v[eids[:test_size]]
        valid_pos_u, valid_pos_v = u[eids[test_size:test_size + valid_size]], v[eids[test_size:test_size + valid_size]]
        train_pos_u, train_pos_v = u[eids[test_size + valid_size:]], v[eids[test_size + valid_size:]]

        # Find all negative edges and split them for training and testing
        adj = sp.coo_matrix((np.ones(len(u)), (u.numpy(), v.numpy())))
        adj_neg = 1 - adj.todense() - np.eye(g.number_of_nodes())
        neg_u, neg_v = np.where(adj_neg != 0)

        neg_eids = np.random.choice(len(neg_u), g.number_of_edges())
        test_neg_u, test_neg_v = neg_u[neg_eids[:test_size]], neg_v[neg_eids[:test_size]]
        valid_neg_u, valid_neg_v = neg_u[neg_eids[test_size:test_size + valid_size]], neg_v[neg_eids[test_size:test_size + valid_size]]
        train_neg_u, train_neg_v = neg_u[neg_eids[test_size + valid_size:]], neg_v[neg_eids[test_size + valid_size:]]

        train_g = dgl.remove_edges(g, eids[:test_size + valid_size])

        train_pos_g = dgl.graph((train_pos_u, train_pos_v), num_nodes=g.number_of_nodes())
        train_neg_g = dgl.graph((train_neg_u, train_neg_v), num_nodes=g.number_of_nodes())

        valid_pos_g = dgl.graph((valid_pos_u, valid_pos_v), num_nodes=g.number_of_nodes())
        valid_neg_g = dgl.graph((valid_neg_u, valid_neg_v), num_nodes=g.number_of_nodes())

        test_pos_g = dgl.graph((test_pos_u, test_pos_v), num_nodes=g.number_of_nodes())
        test_neg_g = dgl.graph((test_neg_u, test_neg_v), num_nodes=g.number_of_nodes())

        return (
            train_g, train_pos_g, train_neg_g,
            valid_pos_g, valid_neg_g, test_pos_g, test_neg_g
        )

    @classmethod
    def __split_edges_train_test(
            cls, g: dgl.DGLGraph, train_ratio: float
    ) -> _typing.Tuple[
        dgl.DGLGraph, dgl.DGLGraph, dgl.DGLGraph,
        dgl.DGLGraph, dgl.DGLGraph,
    ]:
        u, v = g.edges()

        eids = np.arange(g.number_of_edges())
        eids = np.random.permutation(eids)
        test_size = int(len(eids) * (1 - train_ratio))
        train_size = g.number_of_edges() - test_size
        test_pos_u, test_pos_v = u[eids[:test_size]], v[eids[:test_size]]
        train_pos_u, train_pos_v = u[eids[test_size:]], v[eids[test_size:]]

        # Find all negative edges and split them for training and testing
        adj = sp.coo_matrix((np.ones(len(u)), (u.numpy(), v.numpy())))
        adj_neg = 1 - adj.todense() - np.eye(g.number_of_nodes())
        neg_u, neg_v = np.where(adj_neg != 0)

        neg_eids = np.random.choice(len(neg_u), g.number_of_edges())
        test_neg_u, test_neg_v = neg_u[neg_eids[:test_size]], neg_v[neg_eids[:test_size]]
        train_neg_u, train_neg_v = neg_u[neg_eids[test_size:]], neg_v[neg_eids[test_size:]]

        train_g = dgl.remove_edges(g, eids[:test_size])

        train_pos_g = dgl.graph((train_pos_u, train_pos_v), num_nodes=g.number_of_nodes())
        train_neg_g = dgl.graph((train_neg_u, train_neg_v), num_nodes=g.number_of_nodes())

        test_pos_g = dgl.graph((test_pos_u, test_pos_v), num_nodes=g.number_of_nodes())
        test_neg_g = dgl.graph((test_neg_u, test_neg_v), num_nodes=g.number_of_nodes())

        return train_g, train_pos_g, train_neg_g, test_pos_g, test_neg_g

    @classmethod
    def split_edges_for_dgl_graph(
            cls, graph: dgl.DGLGraph,
            train_ratio: float, val_ratio: _typing.Optional[float] = ...
    ) -> _typing.Union[
        _typing.Tuple[
            dgl.DGLGraph, dgl.DGLGraph,
            dgl.DGLGraph, dgl.DGLGraph,
            dgl.DGLGraph, dgl.DGLGraph,
            dgl.DGLGraph
        ],
        _typing.Tuple[
            dgl.DGLGraph, dgl.DGLGraph, dgl.DGLGraph,
            dgl.DGLGraph, dgl.DGLGraph,
        ]
    ]:
        if not 0 < train_ratio < 1:
            raise ValueError(f"Invalid train_ratio as {train_ratio}")
        if isinstance(val_ratio, float):
            if not 0 < val_ratio < 1:
                raise ValueError(f"Invalid val_ratio as {val_ratio}")
            if not 0 < train_ratio + val_ratio < 1:
                raise ValueError(
                    f"Invalid combination (train_ratio, val_ratio) "
                    f"as ({train_ratio}, {val_ratio})"
                )
            return cls.__split_edges_train_val_test(graph, train_ratio, val_ratio)
        else:
            return cls.__split_edges_train_test(graph, train_ratio)


def split_edges_for_data(
        data: _typing.Union[dgl.DGLGraph, autogl.data.graph.GeneralStaticGraph],
        train_ratio: float, val_ratio: _typing.Optional[float] = ...
) -> _typing.Union[
    _typing.Tuple[
        dgl.DGLGraph, dgl.DGLGraph,
        dgl.DGLGraph, dgl.DGLGraph,
        dgl.DGLGraph, dgl.DGLGraph,
        dgl.DGLGraph
    ],
    _typing.Tuple[
        dgl.DGLGraph, dgl.DGLGraph, dgl.DGLGraph,
        dgl.DGLGraph, dgl.DGLGraph,
    ]
]:
    if isinstance(data, dgl.DGLGraph):
        if not data.is_homogeneous:
            raise ValueError(
                "provided DGL graph to split edges MUST be homogeneous"
            )
        else:
            return _SplitEdgesDGLImpl.split_edges_for_dgl_graph(
                data, train_ratio, val_ratio
            )
    elif isinstance(data, autogl.data.graph.GeneralStaticGraph):
        if not (data.nodes.is_homogeneous and data.edges.is_homogeneous):
            raise ValueError(
                "Provided instance of GeneralStaticGraph MUST be homogeneous"
            )
        else:
            return _SplitEdgesDGLImpl.split_edges_for_dgl_graph(
                general_static_graph_to_dgl_graph(data), train_ratio, val_ratio
            )
    else:
        raise TypeError(f"Illegal provided data {data}")



import torch
import torch.nn.functional
import typing as _typing

from torch_geometric.nn.conv import GCNConv
import autogl.data
from . import register_model
from .base import BaseAutoModel, activate_func, ClassificationSupportedSequentialModel
from ....utils import get_logger

LOGGER = get_logger("GCNModel")


class GCN(ClassificationSupportedSequentialModel):
    class _GCNLayer(torch.nn.Module):
        def __init__(
            self,
            input_channels: int,
            output_channels: int,
            add_self_loops: bool = True,
            normalize: bool = True,
            activation_name: _typing.Optional[str] = ...,
            dropout_probability: _typing.Optional[float] = ...,
        ):
            super().__init__()
            self._convolution: GCNConv = GCNConv(
                input_channels,
                output_channels,
                add_self_loops=bool(add_self_loops),
                normalize=bool(normalize),
            )
            if (
                activation_name is not Ellipsis
                and activation_name is not None
                and type(activation_name) == str
            ):
                self._activation_name: _typing.Optional[str] = activation_name
            else:
                self._activation_name: _typing.Optional[str] = None
            if (
                dropout_probability is not Ellipsis
                and dropout_probability is not None
                and type(dropout_probability) == float
            ):
                if dropout_probability < 0:
                    dropout_probability = 0
                if dropout_probability > 1:
                    dropout_probability = 1
                self._dropout: _typing.Optional[torch.nn.Dropout] = torch.nn.Dropout(
                    dropout_probability
                )
            else:
                self._dropout: _typing.Optional[torch.nn.Dropout] = None

        def forward(self, data, enable_activation: bool = True) -> torch.Tensor:
            x: torch.Tensor = getattr(data, "x")
            edge_index: torch.LongTensor = getattr(data, "edge_index")
            edge_weight: _typing.Optional[torch.Tensor] = getattr(
                data, "edge_weight", None
            )
            """ Validate the arguments """
            if not type(x) == type(edge_index) == torch.Tensor:
                raise TypeError
            if edge_weight is not None and (
                type(edge_weight) != torch.Tensor
                or edge_index.size() != (2, edge_weight.size(0))
            ):
                edge_weight: _typing.Optional[torch.Tensor] = None
            x: torch.Tensor = self._convolution.forward(x, edge_index, edge_weight)
            if self._activation_name is not None and enable_activation:
                x: torch.Tensor = activate_func(x, self._activation_name)
            if self._dropout is not None:
                x: torch.Tensor = self._dropout.forward(x)
            return x

    def __init__(
        self,
        num_features: int,
        num_classes: int,
        hidden_features: _typing.Sequence[int],
        activation_name: str,
        dropout: _typing.Union[
            _typing.Optional[float], _typing.Sequence[_typing.Optional[float]]
        ] = None,
        add_self_loops: bool = True,
        normalize: bool = True,
    ):
        if isinstance(dropout, _typing.Sequence):
            if len(dropout) != len(hidden_features) + 1:
                raise TypeError(
                    "When the dropout argument is a sequence, "
                    "The sequence length must equal to the number of layers to construct."
                )
            for _dropout in dropout:
                if _dropout is not None and type(_dropout) != float:
                    raise TypeError(
                        "When the dropout argument is a sequence, "
                        "every item in the sequence must be float or None"
                    )
            dropout_list: _typing.Sequence[_typing.Optional[float]] = dropout
        elif type(dropout) == float:
            if dropout < 0:
                dropout = 0
            if dropout > 1:
                dropout = 1
            dropout_list: _typing.Sequence[_typing.Optional[float]] = [
                dropout for _ in range(len(hidden_features))
            ] + [None]
        elif dropout in (None, Ellipsis, ...):
            dropout_list: _typing.Sequence[_typing.Optional[float]] = [
                None for _ in range(len(hidden_features) + 1)
            ]
        else:
            raise TypeError(
                "The provided dropout argument must be a float number or None or "
                "a sequence in which each item is either a float Number or None."
            )
        super().__init__()
        if len(hidden_features) == 0:
            self.__sequential_encoding_layers: torch.nn.ModuleList = (
                torch.nn.ModuleList(
                    (
                        self._GCNLayer(
                            num_features,
                            num_classes,
                            add_self_loops,
                            normalize,
                            dropout_probability=dropout_list[0],
                        ),
                    )
                )
            )
        else:
            self.__sequential_encoding_layers: torch.nn.ModuleList = (
                torch.nn.ModuleList()
            )
            self.__sequential_encoding_layers.append(
                self._GCNLayer(
                    num_features,
                    hidden_features[0],
                    add_self_loops,
                    normalize,
                    activation_name,
                    dropout_list[0],
                )
            )
            for hidden_feature_index in range(len(hidden_features)):
                if hidden_feature_index + 1 < len(hidden_features):
                    self.__sequential_encoding_layers.append(
                        self._GCNLayer(
                            hidden_features[hidden_feature_index],
                            hidden_features[hidden_feature_index + 1],
                            add_self_loops,
                            normalize,
                            activation_name,
                            dropout_list[hidden_feature_index + 1],
                        )
                    )
                    # print((
                    #         hidden_features[hidden_feature_index],
                    #         hidden_features[hidden_feature_index + 1],
                    #         add_self_loops,
                    #         normalize,
                    #         activation_name,
                    #         dropout_list[hidden_feature_index + 1],
                    #     ))
                else:
                    self.__sequential_encoding_layers.append(
                        self._GCNLayer(
                            hidden_features[hidden_feature_index],
                            num_classes,
                            add_self_loops,
                            normalize,
                            dropout_list[-1],
                        )
                    )
                    # print((
                    #         hidden_features[hidden_feature_index],
                    #         num_classes,
                    #         add_self_loops,
                    #         normalize,
                    #         dropout_list[-1],
                    #     ))

    @property
    def sequential_encoding_layers(self) -> torch.nn.ModuleList:
        return self.__sequential_encoding_layers

    def __extract_edge_indexes_and_weights(
        self, data
    ) -> _typing.Union[
        _typing.Sequence[
            _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]
        ],
        _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]],
    ]:
        def __compose_edge_index_and_weight(
            _edge_index: torch.LongTensor,
            _edge_weight: _typing.Optional[torch.Tensor] = None,
        ) -> _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]:
            if type(_edge_index) != torch.Tensor or _edge_index.dtype != torch.int64:
                print(type(_edge_index))
                raise TypeError
            if _edge_weight is not None and (
                type(_edge_weight) != torch.Tensor
                or _edge_index.size() != (2, _edge_weight.size(0))
            ):
                _edge_weight: _typing.Optional[torch.Tensor] = None
            return _edge_index, _edge_weight

        if not (
            hasattr(data, "edge_indexes")
            and isinstance(getattr(data, "edge_indexes"), _typing.Sequence)
            and len(getattr(data, "edge_indexes"))
            == len(self.__sequential_encoding_layers)
        ):
            return __compose_edge_index_and_weight(
                getattr(data, "edge_index"), getattr(data, "edge_weight", None)
            )
        for __edge_index in getattr(data, "edge_indexes"):
            if type(__edge_index) != torch.Tensor or __edge_index.dtype != torch.int64:
                return __compose_edge_index_and_weight(
                    getattr(data, "edge_index"), getattr(data, "edge_weight", None)
                )

        if (
            hasattr(data, "edge_weights")
            and isinstance(getattr(data, "edge_weights"), _typing.Sequence)
            and len(getattr(data, "edge_weights"))
            == len(self.__sequential_encoding_layers)
        ):
            return [
                __compose_edge_index_and_weight(_edge_index, _edge_weight)
                for _edge_index, _edge_weight in zip(
                    getattr(data, "edge_indexes"), getattr(data, "edge_weights")
                )
            ]
        else:
            return [
                __compose_edge_index_and_weight(__edge_index)
                for __edge_index in getattr(data, "edge_indexes")
            ]

    def forward(self, data):
        return self.cls_decode(self.cls_encode(data))

    def cls_encode(self, data) -> torch.Tensor:
        edge_indexes_and_weights: _typing.Union[
            _typing.Sequence[
                _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]
            ],
            _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]],
        ] = self.__extract_edge_indexes_and_weights(data)

        if (not isinstance(edge_indexes_and_weights, tuple)) and isinstance(
            edge_indexes_and_weights[0], tuple
        ):
            """ edge_indexes_and_weights is sequence of (edge_index, edge_weight) """
            assert len(edge_indexes_and_weights) == len(
                self.__sequential_encoding_layers
            )
            x: torch.Tensor = getattr(data, "x")
            for _edge_index_and_weight, gcn in zip(
                edge_indexes_and_weights, self.__sequential_encoding_layers
            ):
                _temp_data = autogl.data.Data(x=x, edge_index=_edge_index_and_weight[0])
                _temp_data.edge_weight = _edge_index_and_weight[1]
                x = gcn(_temp_data)
            return x
        else:
            """ edge_indexes_and_weights is (edge_index, edge_weight) """
            x = getattr(data, "x")
            for gcn in self.__sequential_encoding_layers:
                _temp_data = autogl.data.Data(
                    x=x, edge_index=edge_indexes_and_weights[0]
                )
                _temp_data.edge_weight = edge_indexes_and_weights[1]
                x = gcn(_temp_data)
            return x

    def cls_decode(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.log_softmax(x, dim=1)

    def lp_encode(self, data):
        x: torch.Tensor = getattr(data, "x")
        for i in range(len(self.__sequential_encoding_layers) - 2):
            x = self.__sequential_encoding_layers[i](
                autogl.data.Data(x, getattr(data, "edge_index"))
            )
        # _GCNLayer(
        #     (_convolution): GCNConv(1433, 128)
        #     (_dropout): Dropout(p=0.0, inplace=False)
        #     )
        # print(self.__sequential_encoding_layers)
        x = self.__sequential_encoding_layers[-2](
            autogl.data.Data(x, getattr(data, "edge_index")), enable_activation=False
        )
        return x

    def lp_decode(self, z, pos_edge_index, neg_edge_index):
        edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
        logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
        return logits

    def lp_decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()


@register_model("gcn-model")
class AutoGCN(BaseAutoModel):
    r"""
    AutoGCN.
    The model used in this automodel is GCN, i.e., the graph convolutional network from the
    `"Semi-supervised Classification with Graph Convolutional
    Networks" <https://arxiv.org/abs/1609.02907>`_ paper. The layer is

    .. math::

        \mathbf{X}^{\prime} = \mathbf{\hat{D}}^{-1/2} \mathbf{\hat{A}}
        \mathbf{\hat{D}}^{-1/2} \mathbf{X} \mathbf{\Theta},

    where :math:`\mathbf{\hat{A}} = \mathbf{A} + \mathbf{I}` denotes the
    adjacency matrix with inserted self-loops and
    :math:`\hat{D}_{ii} = \sum_{j=0} \hat{A}_{ij}` its diagonal degree matrix.

    Parameters
    ----------
    num_features: ``int``
        The dimension of features.

    num_classes: ``int``
        The number of classes.

    device: ``torch.device`` or ``str``
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.
    """

    def __init__(
        self,
        num_features: int = ...,
        num_classes: int = ...,
        device: _typing.Union[str, torch.device] = ...,
        **kwargs
    ) -> None:
        super().__init__(num_features, num_classes, device, **kwargs)
        self.hyper_parameter_space = [
            {
                "parameterName": "add_self_loops",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "normalize",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [128, 64],
            "dropout": 0,
            "act": "relu",
        }

    def _initialize(self):
        self._model = GCN(
            self.input_dimension,
            self.output_dimension,
            self.hyper_parameters.get("hidden"),
            self.hyper_parameters.get("act"),
            self.hyper_parameters.get("dropout", None),
            bool(self.hyper_parameters.get("add_self_loops", True)),
            bool(self.hyper_parameters.get("normalize", True)),
        ).to(self.device)
        # print((
        #     self.input_dimension,
        #     self.output_dimension,
        #     self.hyper_parameters.get("hidden"),
        #     self.hyper_parameters.get("act"),
        #     self.hyper_parameters.get("dropout", None),
        #     bool(self.hyper_parameters.get("add_self_loops", True)),
        #     bool(self.hyper_parameters.get("normalize", True))))



import typing as _typing


class _UniversalRegistryUtility:
    @classmethod
    def to_unique_identifier(cls, name: str) -> str:
        if not isinstance(name, str):
            raise TypeError
        import re
        return re.sub("[^A-Za-z0-9]", '', name.strip()).lower()


class _UniversalRegistryMetaclass(type):
    def __getitem__(cls, k: str):
        __identifier: str = _UniversalRegistryUtility.to_unique_identifier(k)
        if __identifier in cls.__universal_registry:
            return cls.__universal_registry[__identifier][0]
        else:
            raise KeyError(k)

    def __setitem__(cls, k: str, v: type) -> None:
        if not (isinstance(k, str) and isinstance(v, type)):
            raise TypeError
        __identifier: str = _UniversalRegistryUtility.to_unique_identifier(k)
        if __identifier in cls.__universal_registry:
            if not (
                    v == cls.__universal_registry[__identifier][0] and
                    id(v) == id(cls.__universal_registry[__identifier][0])
            ):
                raise ValueError
            cls.__universal_registry[__identifier][1].add(k)
        else:
            cls.__universal_registry[__identifier] = (v, {k})

    def __delitem__(cls, k: str) -> None:
        if not isinstance(k, str):
            raise TypeError
        __identifier: str = _UniversalRegistryUtility.to_unique_identifier(k)
        if __identifier in cls.__universal_registry:
            if k in cls.__universal_registry[__identifier][1]:
                cls.__universal_registry[__identifier][1].remove(k)
            if len(cls.__universal_registry[__identifier][1]) == 0:
                del cls.__universal_registry[__identifier]

    def __contains__(cls, item):
        if isinstance(item, str):
            __identifier: str = _UniversalRegistryUtility.to_unique_identifier(item)
            return __identifier in cls.__universal_registry
        return False

    def __iter__(cls) -> _typing.Iterator[str]:
        results: _typing.MutableSequence[str] = []
        for __identifier, (_, names) in cls.__universal_registry.items():
            results.extend(names)
        return iter(results)

    def __new__(
            mcs, name: str, bases: _typing.Tuple[type, ...],
            namespace: _typing.Dict[str, _typing.Any]
    ):
        return super(_UniversalRegistryMetaclass, mcs).__new__(
            mcs, name, bases, namespace
        )

    def __init__(
            cls, name: str, bases: _typing.Tuple[type, ...],
            namespace: _typing.Dict[str, _typing.Any]
    ):
        super(_UniversalRegistryMetaclass, cls).__init__(
            name, bases, namespace
        )
        cls.__universal_registry: _typing.MutableMapping[
            str, _typing.Tuple[type, _typing.MutableSet[str]]
        ] = {}


class UniversalRegistryBase(metaclass=_UniversalRegistryMetaclass):
    @classmethod
    def to_unique_identifier(cls, name: str) -> str:
        return _UniversalRegistryUtility.to_unique_identifier(name)



import typing as _typing
import torch.utils.data
import torch_geometric
from .target_dependant_sampler import TargetDependantSampler, TargetDependantSampledData


class NeighborSampler(TargetDependantSampler, _typing.Iterable):
    """
    The neighbor sampler from the `"Inductive Representation Learning on
    Large Graphs" <https://arxiv.org/abs/1706.02216>`_ literature, which allows
    for mini-batch training of GNNs on large-scale graphs where full-batch
    training is not feasible.

    Arguments
    ------------
    edge_index:
        A :obj:`torch.LongTensor` that defines the underlying graph
        connectivity/message passing flow.
        :obj:`edge_index` holds the indices of a (sparse) adjacency matrix.
        If :obj:`edge_index` is of type :obj:`torch.LongTensor`, its shape
        must be defined as :obj:`[2, num_edges]`, where messages from nodes
        :obj:`edge_index[0]` are sent to nodes in :obj:`edge_index[1]`
        (in case :obj:`flow="source_to_target"`).
    target_nodes_indexes:
        indexes of target nodes to learn representation.
    sampling_sizes:
        The number of neighbors to sample for each node in each layer.
        If set to :obj:`sampling_sizes[l] = -1`, all neighbors are included in layer :obj:`l`.
    batch_size:
        number of target nodes for each mini-batch.
    num_workers:
        num_workers argument for inner :class:`torch.utils.data.DataLoader`
    shuffle:
        whether to shuffle target nodes for mini-batches.
    """

    class _SequenceDataset(torch.utils.data.Dataset):
        def __init__(self, sequence):
            self.__sequence = sequence

        def __len__(self):
            return len(self.__sequence)

        def __getitem__(self, idx):
            return self.__sequence[idx]

    @classmethod
    def __compute_edge_weight(cls, edge_index: torch.LongTensor) -> torch.Tensor:
        __num_nodes = max(int(edge_index[0].max()), int(edge_index[1].max())) + 1
        __out_degree: torch.LongTensor = torch_geometric.utils.degree(
            edge_index[0], __num_nodes
        )
        __in_degree: torch.LongTensor = torch_geometric.utils.degree(
            edge_index[1], __num_nodes
        )
        temp_tensor: torch.Tensor = torch.stack(
            [__out_degree[edge_index[0]], __in_degree[edge_index[1]]]
        )
        temp_tensor: torch.Tensor = torch.pow(temp_tensor, -0.5)
        temp_tensor[torch.isinf(temp_tensor)] = 0.0
        return temp_tensor[0] * temp_tensor[1]

    def __init__(
        self,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        sampling_sizes: _typing.Sequence[int],
        batch_size: int = 1,
        num_workers: int = 0,
        shuffle: bool = True,
        **kwargs
    ):
        def is_deterministic(__cached: bool = bool(kwargs.get("cached", True))) -> bool:
            if not __cached:
                return False
            _deterministic: bool = True
            for _sampling_size in sampling_sizes:
                if type(_sampling_size) != int:
                    raise TypeError(
                        "The sampling_sizes argument must be a sequence of integer"
                    )
                if _sampling_size >= 0:
                    _deterministic = False
                    break
            return _deterministic

        self.__edge_weight: torch.Tensor = self.__compute_edge_weight(edge_index)
        if int(torch_geometric.__version__.split('.')[0]) > 1:
            from torch_geometric.loader import NeighborSampler as NS
        else:
            from torch_geometric.data import NeighborSampler as NS
        self.__pyg_neighbor_sampler: NS = (
            NS(
                edge_index,
                list(sampling_sizes[::-1]),
                target_nodes_indexes,
                transform=self._transform,
                batch_size=batch_size,
                num_workers=num_workers,
                shuffle=shuffle,
                **kwargs
            )
        )

        if is_deterministic():
            pyg_neighbor_sampler: _typing.Iterable = self.__pyg_neighbor_sampler
            self.__cached_sampled_data_list: _typing.Optional[
                _typing.List[TargetDependantSampledData]
            ] = [sampled_data for sampled_data in pyg_neighbor_sampler]
        else:
            self.__cached_sampled_data_list: _typing.Optional[
                _typing.List[TargetDependantSampledData]
            ] = None

    def _transform(
        self,
        batch_size: int,
        n_id: torch.LongTensor,
        adj_or_adj_list: _typing.Union[
            _typing.Sequence[
                _typing.Tuple[
                    torch.LongTensor, torch.LongTensor, _typing.Tuple[int, int]
                ]
            ],
            _typing.Tuple[torch.LongTensor, torch.LongTensor, _typing.Tuple[int, int]],
        ],
    ) -> TargetDependantSampledData:
        if (
            isinstance(adj_or_adj_list[0], _typing.Tuple)
            and isinstance(adj_or_adj_list, _typing.Sequence)
            and not isinstance(adj_or_adj_list, _typing.Tuple)
        ):
            return TargetDependantSampledData(
                [
                    (
                        current_layer[0],
                        current_layer[1],
                        self.__edge_weight[current_layer[1]],
                    )
                    for current_layer in adj_or_adj_list
                ],
                (torch.arange(batch_size, dtype=torch.long).long(), n_id[:batch_size]),
                n_id,
            )
        elif (
            isinstance(adj_or_adj_list, _typing.Tuple)
            and type(adj_or_adj_list[0]) == torch.Tensor
        ):
            adj_or_adj_list: _typing.Tuple[
                torch.LongTensor, torch.LongTensor, _typing.Tuple[int, int]
            ] = adj_or_adj_list
            return TargetDependantSampledData(
                [
                    (
                        adj_or_adj_list[0],
                        adj_or_adj_list[1],
                        self.__edge_weight[adj_or_adj_list[1]],
                    )
                ],
                (torch.arange(batch_size, dtype=torch.long).long(), n_id[:batch_size]),
                n_id,
            )

    def __iter__(self):
        if self.__cached_sampled_data_list is not None and isinstance(
            self.__cached_sampled_data_list, _typing.Sequence
        ):
            return iter(
                torch.utils.data.DataLoader(
                    self._SequenceDataset(self.__cached_sampled_data_list),
                    collate_fn=lambda x: x[0],
                )
            )
        else:
            return iter(self.__pyg_neighbor_sampler)

    @classmethod
    def create_basic_sampler(
        cls,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        layer_wise_arguments: _typing.Sequence,
        batch_size: int = 1,
        num_workers: int = 1,
        shuffle: bool = True,
        *args,
        **kwargs
    ) -> TargetDependantSampler:
        """
        A static factory method to create instance of :class:`NeighborSampler`

        Arguments
        ------------
        edge_index:
            A :obj:`torch.LongTensor` that defines the underlying graph
            connectivity/message passing flow.
            :obj:`edge_index` holds the indices of a (sparse) adjacency matrix.
            If :obj:`edge_index` is of type :obj:`torch.LongTensor`, its shape
            must be defined as :obj:`[2, num_edges]`, where messages from nodes
            :obj:`edge_index[0]` are sent to nodes in :obj:`edge_index[1]`
            (in case :obj:`flow="source_to_target"`).
        target_nodes_indexes:
            indexes of target nodes to learn representation.
        layer_wise_arguments:
            The number of neighbors to sample for each node in each layer.
            If set to :obj:`sampling_sizes[l] = -1`, all neighbors are included in layer :obj:`l`.
        batch_size:
            number of target nodes for each mini-batch.
        num_workers:
            num_workers argument for inner :class:`torch.utils.data.DataLoader`
        shuffle:
            whether to shuffle target nodes for mini-batches.
        """
        return cls(
            edge_index,
            target_nodes_indexes,
            layer_wise_arguments,
            batch_size,
            num_workers,
            shuffle,
            **kwargs
        )
